"""Class to handle project settings"""
import configparser
import platform
from os import environ
from pathlib import Path


class LastShoutSettings:
    """Class to handle project settings"""

    @staticmethod
    def get_settings_folder():
        """Method to determine where the users settings should be saved"""
        if platform.system() == "Windows":
            folder = Path(environ["APPDATA"]).joinpath("LastShout")
        else:
            folder = Path.home() / ".LastShout"

        return folder

    def set_default_values(self):
        """Method to set default values as empty"""
        # Last.fm credentials
        self.last_user = ""
        self.last_access_key = ""

        # Twitter credentials
        self.consumer_key = ""
        self.consumer_secret = ""
        self.access_key = ""
        self.access_secret = ""

        # Mastodon credentials
        self.mastodon_client_id = ""
        self.mastodon_client_secret = ""
        self.mastodon_user_token = ""
        self.mastodon_api_base_url = ""

    def __init__(self, config_dir=None):
        self.set_default_values()

        self.config = configparser.ConfigParser()
        if config_dir:
            self.folder = Path(config_dir)
        else:
            self.folder = LastShoutSettings.get_settings_folder()

        if not self.folder.is_dir():
            self.folder.mkdir(parents=True)

        self.settings_file = self.folder.joinpath("settings.ini")

        # Write the config file if it doesn't exist
        if not self.settings_file.is_file():
            self.save()
        else:
            self.load()

    def load(self):
        """ Method to retrieve user's settings """
        self.config.read(self.settings_file)

        if self.config.has_option("last_fm", "user"):
            self.last_user = self.config["last_fm"]["user"]

        if self.config.has_option("last_fm", "access_key"):
            self.last_access_key = self.config["last_fm"]["access_key"]

        if self.config.has_option("twitter", "consumer_key"):
            self.consumer_key = self.config["twitter"]["consumer_key"]

        if self.config.has_option("twitter", "consumer_secret"):
            self.consumer_secret = self.config["twitter"]["consumer_secret"]

        if self.config.has_option("twitter", "access_key"):
            self.access_key = self.config["twitter"]["access_key"]

        if self.config.has_option("twitter", "access_secret"):
            self.access_secret = self.config["twitter"]["access_secret"]

        if self.config.has_option("mastodon", "client_id"):
            self.mastodon_client_id = self.config["mastodon"]["client_id"]

        if self.config.has_option("mastodon", "client_secret"):
            self.mastodon_client_secret = self.config["mastodon"]["client_secret"]

        if self.config.has_option("mastodon", "user_token"):
            self.mastodon_user_token = self.config["mastodon"]["user_token"]

        if self.config.has_option("mastodon", "api_base_url"):
            self.mastodon_api_base_url = self.config["mastodon"]["api_base_url"]

    def save(self):
        """ Method to save user's settings """
        if not self.config.has_section("last_fm"):
            self.config.add_section("last_fm")

        if not self.config.has_section("twitter"):
            self.config.add_section("twitter")

        if not self.config.has_section("mastodon"):
            self.config.add_section("mastodon")

        self.config["last_fm"]["user"] = self.last_user
        self.config["last_fm"]["access_key"] = self.last_access_key

        self.config["twitter"]["consumer_key"] = self.consumer_key
        self.config["twitter"]["consumer_secret"] = self.consumer_secret
        self.config["twitter"]["access_key"] = self.access_key
        self.config["twitter"]["access_secret"] = self.access_secret

        self.config["mastodon"]["client_id"] = self.mastodon_client_id
        self.config["mastodon"]["client_secret"] = self.mastodon_client_secret
        self.config["mastodon"]["user_token"] = self.mastodon_user_token
        self.config["mastodon"]["api_base_url"] = self.mastodon_api_base_url

        with self.settings_file.open(mode="w") as configfile:
            self.config.write(configfile)



""" Function to make the parser """
import argparse

from last_shout import VERSION


def create_parser():
    """ Function creating the parser """
    parser = argparse.ArgumentParser(
        description="A program to post last.fm statistics"
        + " to Twitter and/or Mastodon.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("-u", "--user", help="Last.fm username")
    parser.add_argument("--last-access-key", help="Last.fm access key")
    parser.add_argument(
        "--set-lastfm",
        help="Set Last.fm credentials",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-n", "--number", type=int, default=10, help="Number of last.fm top artists",
    )
    parser.add_argument(
        "-p",
        "--period",
        default="7day",
        help="The time period over which to retrieve top artists.\n"
        + "Options are:\n"
        + "overall | 7day | 1month | 3month | 6month | 12month",
    )
    parser.add_argument(
        "-t",
        "--tweet",
        help="Post Last.fm stats to Twitter",
        action="store_true",
        default=False,
    )
    parser.add_argument("--consumer-key", help="Twitter consumer key")
    parser.add_argument("--consumer-secret", help="Twitter consumer secret")
    parser.add_argument("--access-key", help="Twitter access token key")
    parser.add_argument("--access-secret", help="Twitter access secret")
    parser.add_argument(
        "--set-twitter",
        help="Set Twitter credentials",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--create-mastodon-app",
        help="Create mastodon application",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--create-mastodon-user",
        help="Create Mastodon user token",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--toot",
        help="Post Last.fm stats to Mastodon",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(VERSION),
        help="Show the version number and exit",
    )

    return parser



""" Various utilites """
import datetime

MUSICAL_NOTE = "\u266A"


def periods_to_string(period):
    """ Function to convert time period setting to string value """
    now = datetime.datetime.now()
    switcher = {
        "overall": "All-Time",
        "7day": "Weekly",
        "1month": "Monthly",
        "3month": "Quarterly",
        "6month": "Semi-Annual",
        "12month": str(now.year),
    }

    return switcher.get(period, " ")


# TODO: Add function to truncate tweet if it's length is greater than 280 characters.


def build_twitter_string(artists, period):
    """
    Function to convert Last.fm result to a string
    that will be posted to Twitter
    """
    total = len(artists)
    txt = f"{MUSICAL_NOTE} My {periods_to_string(period)} Top {total} #lastfm artists: "

    for count, artist in enumerate(artists, 1):
        txt += f"{artist.item.name} ({artist.weight})"
        if count < total - 1:
            txt += ", "
        elif count == total - 1:
            txt += " & "

    return txt



""" Function to users top artist from Last.fm """
import pylast


def get_top_artist(last_access_key, username, number, period):
    """ Simple function returning the users top artists """
    network = pylast.LastFMNetwork(last_access_key)
    user = network.get_user(username)
    return user.get_top_artists(period=period, limit=number)



""" Function to send tweet to Twitter """
import sys

import twitter


def send_tweet(settings, tweet, encoding):
    """ Simple function to send tweet to Twitter """
    api = twitter.Api(
        consumer_key=settings.consumer_key,
        consumer_secret=settings.consumer_secret,
        access_token_key=settings.access_key,
        access_token_secret=settings.access_secret,
        input_encoding=encoding,
    )

    try:
        status = api.PostUpdate(tweet)
    except UnicodeDecodeError:
        print("Last.fm statistics could not be encoded for Twitter.")
        sys.exit(2)
    except twitter.TwitterError as error:
        print(f"Twitter error: {error.message}")
        sys.exit(2)

    return status



""" Main project file """

import sys

from mastodon import Mastodon, MastodonIllegalArgumentError

from .libshout.lastfm import get_top_artist
from .libshout.options import create_parser
from .libshout.settings import LastShoutSettings
from .libshout.twitter import send_tweet
from .libshout.utils import build_twitter_string

MASTODON_REDIRECT_URI = "urn:ietf:wg:oauth:2.0:oob"


def has_lastfm_credentials(settings):
    return bool(settings.last_user or not settings.last_access_key)


def has_twitter_credentials(settings):
    return bool(
        (
            settings.consumer_key
            and settings.consumer_secret
            and settings.access_key
            and settings.access_secret
        )
    )


def has_mastodon_app_credentials(settings):
    return bool(
        (
            settings.mastodon_client_id
            and settings.mastodon_client_secret
            and settings.mastodon_api_base_url
        )
    )


def has_mastodon_user_credentials(settings):
    return bool(settings.mastodon_user_token)


def create_mastodon_app(settings):
    app_name = "Last-Shout"
    app_url = "https://github.com/bpepple/last-shout"
    instance = input("Enter Mastodon instance (ex. 'https://mastodon.social'): ")

    client_id, client_secret = Mastodon.create_app(
        app_name,
        website=app_url,
        api_base_url=instance,
        redirect_uris=MASTODON_REDIRECT_URI,
    )

    if not (client_id or client_secret):
        return False

    settings.mastodon_client_id = client_id
    settings.mastodon_client_secret = client_secret
    settings.mastodon_api_base_url = instance
    settings.save()

    return True


def create_mastodon_user_token(settings):
    mastodon = Mastodon(
        client_id=settings.mastodon_client_id,
        client_secret=settings.mastodon_client_secret,
        api_base_url=settings.mastodon_api_base_url,
    )
    # Uri to provide to client to grant authorization
    print("Go to the follow url to grant authorization from Mastodon.\n")
    print(
        mastodon.auth_request_url(
            client_id=settings.mastodon_client_id, redirect_uris=MASTODON_REDIRECT_URI,
        )
    )

    auth = input("\nCopy the authorized code here to generate user token: ")
    try:
        user_token = mastodon.log_in(
            code=auth, scopes=["write"], redirect_uri=MASTODON_REDIRECT_URI,
        )
    except MastodonIllegalArgumentError:
        return False

    settings.mastodon_user_token = user_token
    settings.save()

    return True


def sent_toot(settings, toot_text):
    mastodon = Mastodon(
        access_token=settings.mastodon_user_token,
        api_base_url=settings.mastodon_api_base_url,
    )
    return mastodon.toot(toot_text)


def main():
    """ Main Function """
    settings = LastShoutSettings()
    parser = create_parser()
    opts = parser.parse_args()

    # Create Mastodon application
    if opts.create_mastodon_app:
        result = create_mastodon_app(settings)
        if result:
            print("Saved Mastodon application credentials to configuration file.")
        else:
            print("Unable to create application credentials.")
        sys.exit(0)

    # Create Mastodon user token
    if opts.create_mastodon_user:
        if not has_mastodon_app_credentials(settings):
            print("Missing Mastodon application credentials. Exiting...")
            sys.exit(2)
        result = create_mastodon_user_token(settings)
        if result:
            print("Saved Mastodon user token to configuration file.")
        else:
            print("Unable to create Mastododon user token.")
        sys.exit(0)

    # Get Last.fm credentials
    if opts.user:
        settings.last_user = opts.user

    if opts.last_access_key:
        settings.last_access_key = opts.last_access_key

    # Get Twitter credentials
    if opts.consumer_key:
        settings.consumer_key = opts.consumer_key

    if opts.consumer_secret:
        settings.consumer_secret = opts.consumer_secret

    if opts.access_key:
        settings.access_key = opts.access_key

    if opts.access_secret:
        settings.access_secret = opts.access_secret

    # Save Last.fm options
    if opts.set_lastfm:
        if has_lastfm_credentials(settings):
            settings.save()
        else:
            print("Missing Last.fm credetials. Unable to save.")
            sys.exit(0)

    # Save Twitter options
    if opts.set_twitter:
        if has_twitter_credentials(settings):
            settings.save()
        else:
            print("Missing Twitter credentials. Unable to save.")
            sys.exit(0)

    # If Last.fm or Twitter credentials are missing exit
    if not has_lastfm_credentials(settings) or not has_twitter_credentials(settings):
        print("Missing Last.fm or Twitter credentials. Exitting...")
        sys.exit(2)

    artists = get_top_artist(
        settings.last_access_key, settings.last_user, opts.number, opts.period
    )
    twitter_text = build_twitter_string(artists, opts.period)

    if opts.tweet:
        status = send_tweet(settings, twitter_text, None)
        print(f"Last.fm statistics posted to Twitter at {status.created_at}")

    if opts.toot:
        if not has_mastodon_app_credentials(
            settings
        ) or not has_mastodon_user_credentials(settings):
            print("Missing Mastodon credentials. Exiting...")
            sys.exit(2)

        status = sent_toot(settings, twitter_text)
        print(f"Last.fm statistics posted to Mastodon at {status.created_at}")

    if not opts.tweet and not opts.toot:
        print(twitter_text)


if __name__ == "__main__":
    sys.exit(main())
